#include "Eigen/Dense"
#include <iostream>
#include <cassert>

template<typename T, int M, int N>
Eigen::Matrix<T,1,N> dEdp(const Eigen::Matrix<T,N,1> &p, const Eigen::Matrix<T,M,N> &x, const Eigen::Matrix<T,M,1> &y) {
  auto m=y.size(), n=p.size();
  Eigen::Matrix<T,1,N> r=Eigen::Matrix<T,1,N>::Zero(1,n);
  for (auto i=0;i<m;i++) r+=(x.row(i).dot(p)-y(i))*x.row(i);
  return 2*r;
}

template<typename T, int M, int N>
Eigen::Matrix<T,N,N> ddEdpp(const Eigen::Matrix<T,N,1> &p, const Eigen::Matrix<T,M,N> &x, const Eigen::Matrix<T,M,1> &y) {
  auto m=y.size(), n=p.size();
  Eigen::Matrix<T,N,N> r=Eigen::Matrix<T,N,N>::Zero(n,n);
  for (auto i=0;i<m;i++) r+=x.row(i).transpose()*x.row(i);
  return 2*r;
}

template<typename T, int M, int N>
void Givens(const Eigen::Matrix<T,M,N> &A, Eigen::Matrix<T,M,M> &QT, Eigen::Matrix<T,M,N> &R) {
  auto m=A.rows(), n=A.cols();
  using MT=Eigen::Matrix<T,M,M>;
  R=A; QT=MT::Identity(m,m);
  for (int j=0;j<n;j++) {
    for (int i=m-2;i>=j;i--) {
      T norm_a_tilde=R.col(j).block(i,0,2,1).norm();
      MT G=MT::Identity(m,m);
      G(i,i)=R.col(j)(i)/norm_a_tilde; G(i+1,i+1)=G(i,i);
      G(i,i+1)=R.col(j)(i+1)/norm_a_tilde; G(i+1,i)=-G(i,i+1);
      QT=G*QT; R=G*R;
    }   
  }
}

template<typename T, int M, int N>
void LinearSolve(const Eigen::Matrix<T,M,M> &QT, const Eigen::Matrix<T,M,N> &R, Eigen::Matrix<T,N,1> &p, const Eigen::Matrix<T,M,1> &y) {
  auto n=R.cols();
  Eigen::Matrix<T,M,1> r=QT*y;
  for (auto i=n-1;i>=0;i--) {
    T d=r(i);
    for (auto j=n-1;j>i;j--) d-=R(i,j)*p(j);
    p(i)=d/R(i,i);
  }
}

template<typename T, int M, int N>
void Regression(Eigen::Matrix<T,N,1>& p, const Eigen::Matrix<T,M,N> &x, const Eigen::Matrix<T,M,1> &y) {
  auto m=x.rows(), n=x.cols();
  Eigen::Matrix<T,M,M> QT(m,m); Eigen::Matrix<T,M,N> R(m,n);
  Givens(x,QT,R);
  LinearSolve(QT,R,p,y);
}

int main(int argc, char* argv[]) {
  assert(argc==3); auto m=std::stoi(argv[1]), n=std::stoi(argv[2]);
  using T=double;
  using MT=Eigen::Matrix<T,Eigen::Dynamic,Eigen::Dynamic>;
  using VT=Eigen::Matrix<T,Eigen::Dynamic,1>;
  MT x=MT::Random(m,n); x=x.cwiseProduct(x);
  VT p(n), y=VT::Random(m);
  Regression(p,x,y);
  std::cout << "p^T=" << p.transpose() << std::endl;
  std::cout << "E'=" << dEdp(p,x,y) << std::endl;
  std::cout << "E''=" << std::endl << ddEdpp(p,x,y) << std::endl;
  Eigen::LLT<Eigen::Matrix<T,Eigen::Dynamic,Eigen::Dynamic>> spd_test(ddEdpp(p,x,y));
  if (spd_test.info()!=Eigen::NumericalIssue)
    std::cout << "Hessian is spd." << std::endl;
  else
    std::cout << "Hessian appears to be not spd!" << std::endl;
  return 0;
}
