#include "Eigen/Dense"
#include <iostream>
#include <cassert>
#include <fstream>

template<typename T, int M, int N>
T E(const Eigen::Matrix<T,N,1> &p, const Eigen::Matrix<T,M,N> &x, const Eigen::Matrix<T,M,1> &y) {
  return (x*p-y).squaredNorm();
}

template<typename T, int M, int N>
Eigen::Matrix<T,1,N> dEdp(const Eigen::Matrix<T,N,1> &p, const Eigen::Matrix<T,M,N> &x, const Eigen::Matrix<T,M,1> &y) {
  auto m=y.size(), n=p.size();
  Eigen::Matrix<T,1,N> r=Eigen::Matrix<T,1,N>::Zero(1,n);
  for (auto i=0;i<m;i++) r+=(x.row(i).dot(p)-y(i))*x.row(i);
  return 2*r;
}

template<typename T, int M, int N>
Eigen::Matrix<T,N,N> ddEdpp(const Eigen::Matrix<T,N,1> &p, const Eigen::Matrix<T,M,N> &x, const Eigen::Matrix<T,M,1> &y) {
  auto m=y.size(), n=p.size();
  Eigen::Matrix<T,N,N> r=Eigen::Matrix<T,N,N>::Zero(n,n);
  for (auto i=0;i<m;i++) r+=x.row(i).transpose()*x.row(i);
  return 2*r;
}

template<typename T, int M, int N>
void NormalEquation(Eigen::Matrix<T,N,1>& p, const Eigen::Matrix<T,M,N> &x, const Eigen::Matrix<T,M,1> &y) {
  p=(x.transpose()*x).llt().solve(x.transpose()*y);
}

template<typename T, int M, int N>
void Regression(Eigen::Matrix<T,N,1>& p, const Eigen::Matrix<T,M,N> &x, const Eigen::Matrix<T,M,1> &y) {
  NormalEquation(p,x,y);
}

int main(int argc, char* argv[]) {
  assert(argc==3); auto m=std::stoi(argv[1]), n=std::stoi(argv[2]);
  using T=double;
  using MT=Eigen::Matrix<T,Eigen::Dynamic,Eigen::Dynamic>;
  using VT=Eigen::Matrix<T,Eigen::Dynamic,1>;
  MT x=MT::Random(m,n); x=x.cwiseProduct(x);
  VT p(n), y=VT::Random(m);
  if (n<3) {
    std::ofstream plot_data("data");
    for (int i=0;i<m;i++) {
      for (int j=0;j<n;j++) plot_data << x(i,j) << " ";
      plot_data << y(i) << std::endl;
    }
    plot_data.close();
  }
  Regression(p,x,y);
  std::cout << "p^T=" << p.transpose() << std::endl;
  std::cout << "E'=" << dEdp(p,x,y) << std::endl;
  std::cout << "E''=" << std::endl << ddEdpp(p,x,y) << std::endl;
  Eigen::LLT<Eigen::Matrix<T,Eigen::Dynamic,Eigen::Dynamic>> spd_test(ddEdpp(p,x,y));
  if (spd_test.info()!=Eigen::NumericalIssue)
    std::cout << "Hessian is spd." << std::endl;
  else
    std::cout << "Hessian appears to be not spd!" << std::endl;
  return 0;
}
