#include "Eigen/Dense"
#include <iostream>
#include <cassert>

#include "dco.hpp"

template<typename TA, typename TP, int N>
TA f(const Eigen::Matrix<TA,N,1> &p, const Eigen::Matrix<TP,N,1> &x) { 
  int n=x.size();
  TA r=0;
  for (int i=0;i<n;i++) r+=p(i)*x(i); 
  return pow(r,2); 
}

template<typename TA, typename TP, int M, int N>
Eigen::Matrix<TA,M,1> F(const Eigen::Matrix<TA,N,1> &p, const Eigen::Matrix<TP,M,N> &x, const Eigen::Matrix<TP,M,1> &y) {
  int m=y.size();
  Eigen::Matrix<TA,M,1> r(m);
  for (auto i=0;i<m;i++) {
    Eigen::Matrix<TP,N,1> xr=x.row(i);
    r(i)=f(p,xr)-y(i);
  }
  return r;
}

template<typename T, int M, int N>
Eigen::Matrix<T,M,N> dFdp(const Eigen::Matrix<T,N,1> &p_v, const Eigen::Matrix<T,M,N> &x, const Eigen::Matrix<T,M,1> &y) {
  int m=y.size(), n=p_v.size();
  Eigen::Matrix<T,M,N> rj(m,n);
  using DCO_T=typename dco::gt1s<T>::type;
  Eigen::Matrix<DCO_T,N,1> p(n);
  for (int i=0;i<n;i++) p(i)=p_v(i);
  Eigen::Matrix<DCO_T,M,1> r(m);
  for (int i=0;i<n;i++) {
    dco::derivative(p(i))=1;
    r=F(p,x,y);
    for (int k=0;k<m;k++)
      rj(k,i)=dco::derivative(r(k));
    dco::derivative(p(i))=0;
  }
  return rj;
}

template<typename TA, typename TP,int M, int N>
TA E(const Eigen::Matrix<TA,N,1> &p, const Eigen::Matrix<TP,M,N> &x, const Eigen::Matrix<TP,M,1> &y) {
  int m=y.size();
  TA o=0;
  for (auto i=0;i<m;i++) {
    Eigen::Matrix<TP,N,1> xr=x.row(i);
    o+=pow(f(p,xr)-y(i),2);
  }
  return o;
}

template<typename TA, typename TP, int M, int N>
Eigen::Matrix<TA,1,N> dEdp(const Eigen::Matrix<TA,N,1> &p_v, const Eigen::Matrix<TP,M,N> &x, const Eigen::Matrix<TP,M,1> &y) {
  using DCO_M=typename dco::ga1s<TA>;
  using DCO_T=typename DCO_M::type;
  using DCO_TT=typename DCO_M::tape_t;
  int n=p_v.size();
  Eigen::Matrix<TA,1,N> dodp(1,n); 
  Eigen::Matrix<DCO_T,N,1> p(n); DCO_T o=0;
  DCO_M::global_tape=DCO_TT::create();
  for (auto i=0;i<n;i++) {
    p(i)=p_v(i);
    DCO_M::global_tape->register_variable(p(i));
  }
  o=E(p,x,y);
  dco::derivative(o)=1;
  DCO_M::global_tape->interpret_adjoint();
  for (auto i=0;i<n;i++) dodp(0,i)=dco::derivative(p(i));
  DCO_TT::remove(DCO_M::global_tape);
  return dodp;
}

template<typename T, int M, int N>
Eigen::Matrix<T,N,N> ddEdpp(const Eigen::Matrix<T,N,1> &p_v, const Eigen::Matrix<T,M,N> &x, const Eigen::Matrix<T,M,1> &y) {
  using DCO_T=typename dco::gt1s<T>::type;
  int n=p_v.size();
  Eigen::Matrix<T,N,N> ddodpp(n,n);
  Eigen::Matrix<DCO_T,N,1> p(n);
  for (auto i=0;i<n;i++) p(i)=p_v(i);
  for (auto i=0;i<n;i++) {
    dco::derivative(p(i))=1;
    Eigen::Matrix<DCO_T,N,1> dodp=dEdp(p,x,y);
    for (auto j=0;j<n;j++) ddodpp(j,i)=dco::derivative(dodp(j));
    dco::derivative(p(i))=0;
  }
  return ddodpp;
}

template<typename T, int M, int N>
void NormalEquation(const Eigen::Matrix<T,M,N> &A, Eigen::Matrix<T,N,1> &p, const Eigen::Matrix<T,M,1> &y) { 
  p=(A.transpose()*A).llt().solve(-A.transpose()*y);
}

template<typename T, int M, int N>
void Regression(Eigen::Matrix<T,N,1> &p, const Eigen::Matrix<T,M,N> &x, const Eigen::Matrix<T,M,1> &y, const T& eps) {    
  T o=E(p,x,y),o_prev;
  while (dEdp(p,x,y).norm()>eps) {
    o_prev=o;
    Eigen::Matrix<T,N,1> delta_p;
    Eigen::Matrix<T,M,1> r=F(p,x,y); 
    Eigen::Matrix<T,M,N> drdp=dFdp(p,x,y);
    NormalEquation(drdp,delta_p,r);
    T alpha=2.;
    while (o_prev<=o) {
      Eigen::Matrix<T,N,1> p_trial=p;
      alpha/=2;
      p_trial+=alpha*delta_p; o=E(p_trial,x,y);
    }
    p+=alpha*delta_p;
  }
}

int main(int argc, char* argv[]) {
  assert(argc==3); auto m=std::stoi(argv[1]), n=std::stoi(argv[2]);
  using T=double;
  using MT=Eigen::Matrix<T,Eigen::Dynamic,Eigen::Dynamic>;
  using VT=Eigen::Matrix<T,Eigen::Dynamic,1>;
  MT x=MT::Random(m,n); x=x.cwiseProduct(x);
  VT y=VT::Random(m);
  VT p=VT::Random(n); 
  Regression(p,x,y,1e-5);
  std::cout << "p^T=" << p.transpose() << std::endl;
  std::cout << "E'=" << dEdp(p,x,y) << std::endl;
  std::cout << "E''=" << std::endl << ddEdpp(p,x,y) << std::endl;
  Eigen::LLT<Eigen::Matrix<T,Eigen::Dynamic,Eigen::Dynamic>> spd_test(ddEdpp    (p,x,y));
  if (spd_test.info()!=Eigen::NumericalIssue)
    std::cout << "Hessian is spd." << std::endl;
  else
    std::cout << "Hessian appears to be not spd!" << std::endl;
  return 0;
}
