#include "Eigen/Dense"
#include <iostream>
#include <cassert>
#include <limits>

template<typename T, int N>
T f(const Eigen::Matrix<T,N,1> &p, const Eigen::Matrix<T,N,1> &x) { 
  return pow(p.dot(x),2); 
}

template<typename T, int M, int N>
Eigen::Matrix<T,M,1> F(const Eigen::Matrix<T,N,1> &p, const Eigen::Matrix<T,M,N> &x, const Eigen::Matrix<T,M,1> &y) {
  int m=y.size();
  Eigen::Matrix<T,M,1> r(m);
  for (auto i=0;i<m;i++) {
    Eigen::Matrix<T,N,1> xr=x.row(i);
    r(i)=f(p,xr)-y(i);
  }
  return r;
}

template<typename T, int M, int N>
Eigen::Matrix<T,M,N> dFdp(const Eigen::Matrix<T,N,1> &p, const Eigen::Matrix<T,M,N> &x, const Eigen::Matrix<T,M,1> &y) {
  int n=p.size(), m=y.size();
  Eigen::Matrix<T,M,N> rj(m,n);
  for (auto i=0;i<n;i++) {
    T dp=fabs(p(i))<1 ? sqrt(sqrt(std::numeric_limits<T>::epsilon()))
       : sqrt(sqrt(std::numeric_limits<T>::epsilon())*fabs(p(i)));
    Eigen::Matrix<T,N,1> pp=p+dp*Eigen::Matrix<T,N,1>::Unit(n,i);
    Eigen::Matrix<T,N,1> pm=p-dp*Eigen::Matrix<T,N,1>::Unit(n,i);
    Eigen::Matrix<T,M,1> rji=(F(pp,x,y)-F(pm,x,y))/(2*dp);
    for (auto j=0;j<m;j++) rj(j,i)=rji(j);
  }
  return rj;
}

template<typename T, int M, int N>
T E(const Eigen::Matrix<T,N,1> &p, const Eigen::Matrix<T,M,N> &x, const Eigen::Matrix<T,M,1> &y) {
  T o=0;
  for (auto i=0;i<y.size();i++) o+=pow(pow(x.row(i)*p,2)-y(i),2);
  return o; 
} 

template<typename T, int M, int N>
Eigen::Matrix<T,1,N> dEdp(const Eigen::Matrix<T,N,1> &p, const Eigen::Matrix<T,M,N> &x, const Eigen::Matrix<T,M,1> &y) {
  int n=p.size();
  Eigen::Matrix<T,1,N> r(1,n);
  for (auto i=0;i<n;i++) {
    T dp=fabs(p(i))<1 ? sqrt(sqrt(std::numeric_limits<T>::epsilon()))
       : sqrt(sqrt(std::numeric_limits<T>::epsilon())*fabs(p(i)));
    Eigen::Matrix<T,N,1> pp=p+dp*Eigen::Matrix<T,N,1>::Unit(n,i);
    Eigen::Matrix<T,N,1> pm=p-dp*Eigen::Matrix<T,N,1>::Unit(n,i);
    r(0,i)=(E(pp,x,y)-E(pm,x,y))/(2*dp);
  }
  return r;
}

template<typename T, int M, int N>
Eigen::Matrix<T,N,N> ddEdpp(const Eigen::Matrix<T,N,1> &p, const Eigen::Matrix<T,M,N> &x, const Eigen::Matrix<T,M,1> &y) {
  int n=p.size();
  Eigen::Matrix<T,N,N> r(n,n);
  for (auto i=0;i<n;i++) {
    T dp=fabs(p(i))<1 ? sqrt(sqrt(std::numeric_limits<T>::epsilon()))
       : sqrt(sqrt(std::numeric_limits<T>::epsilon())*fabs(p(i)));
    Eigen::Matrix<T,N,1> pp=p+dp*Eigen::Matrix<T,N,1>::Unit(n,i);
    Eigen::Matrix<T,N,1> pm=p-dp*Eigen::Matrix<T,N,1>::Unit(n,i);
    Eigen::Matrix<T,N,1> ri=(dEdp(pp,x,y)-dEdp(pm,x,y))/(2*dp);
    for (auto j=0;j<n;j++) r(j,i)=ri(j);
  }
  return r;
}

template<typename T, int M, int N>
void NormalEquation(const Eigen::Matrix<T,M,N> &A, Eigen::Matrix<T,N,1> &p, const Eigen::Matrix<T,M,1> &y) { 
  p=(A.transpose()*A).llt().solve(-A.transpose()*y);
}

template<typename T, int M, int N>
void Regression(Eigen::Matrix<T,N,1> &p, const Eigen::Matrix<T,M,N> &x, const Eigen::Matrix<T,M,1> &y, const T& eps) {    
  T o=E(p,x,y),o_prev;
  while (dEdp(p,x,y).norm()>eps) {
    o_prev=o;
    Eigen::Matrix<T,N,1> delta_p;
    Eigen::Matrix<T,M,1> r=F(p,x,y); 
    Eigen::Matrix<T,M,N> drdp=dFdp(p,x,y);
    NormalEquation(drdp,delta_p,r);
    T alpha=2.;
    while (o_prev<=o) {
      Eigen::Matrix<T,N,1> p_trial=p;
      alpha/=2;
      p_trial+=alpha*delta_p; o=E(p_trial,x,y);
    }
    p+=alpha*delta_p;
  }
}

int main(int argc, char* argv[]) {
  assert(argc==3); auto m=std::stoi(argv[1]), n=std::stoi(argv[2]);
  using T=double;
  using MT=Eigen::Matrix<T,Eigen::Dynamic,Eigen::Dynamic>;
  using VT=Eigen::Matrix<T,Eigen::Dynamic,1>;
  MT x=MT::Random(m,n); x=x.cwiseProduct(x);
  VT y=VT::Random(m);
  VT p=VT::Random(n); 
  Regression(p,x,y,1e-5);
  std::cout << "p^T=" << p.transpose() << std::endl;
  std::cout << "E'=" << dEdp(p,x,y) << std::endl;
  std::cout << "E''=" << std::endl << ddEdpp(p,x,y) << std::endl;
  Eigen::LLT<Eigen::Matrix<T,Eigen::Dynamic,Eigen::Dynamic>> spd_test(ddEdpp    (p,x,y));
  if (spd_test.info()!=Eigen::NumericalIssue)
    std::cout << "Hessian is spd." << std::endl;
  else
    std::cout << "Hessian appears to be not spd!" << std::endl;
  return 0;
}
